program MillBase3x;

{ $DEFINE LCD}              // LCD is for Debug purpose only
{ $DEFINE MillControl}

{ $NOSHADOW}               // we use all register save for app interrupts

Device = mega128, VCC = 5;

Define_Fuses
  LockBits0      = [];
  FuseBits0      = [BODEN];
  FuseBits1      = [CKOPT, SPIEN, JTAGEN, OCDEN];
  FuseBits2      = [];
  ProgMode       = JTAG;

Import SysTick, SerPort, TWInet, TickTimer, SwitchPort1;
{$IFDEF LCD}
Import LCDmultiPort;
{$ENDIF}

From System Import LongWord, LongInt;

Define
  ProcClock      = 16000000;        {Hertz}
  SysTick        = 10;              {msec}
  StackSize      = $0040, iData;
  FrameSize      = $0100, iData;
  SerPort        = 57600, Stop2;    {Baud, StopBits|Parity}
  RxBuffer       = 40, iData;
  TxBuffer       = 40, iData;
  TickTimer      = Timer3;
  
  TWInetMode     = Master;
  TWInode        = 02;              {default address in slave mode}
  TWIpresc       = TWI_BR400;       {Bitrate}
  TWIframe       = 40, iData;       {buffer/packet size}
  TWIframeBC     = 32;              {Broadcast Frame size}
  
  SwitchPort1    = PinE;

{$IFDEF LCD}
  LCDmultiPort   = I2C_TWI;
  LCDrows_M      = 4;              {rows}
  LCDcolumns_M   = 20;             {columns per line}
  LCDtype_M      = 44780;
{$ENDIF}


Implementation

{$IDATA}

{--------------------------------------------------------------}
{ Type Declarations }

type
  tStr10     = string[10];
  str20      = string[20];
  
  // MillControl
  tCtrlMsg   = (mdIdle, mdJog, mdJogging, mdSpindleOn, mdStart, mdWork,
                mdSpindleOff, mdStop, mdState, mdBusy, mdString, mdParam);

  tCtrlTxRec = record
                 Msg      : tCtrlMsg;
                 line     : byte;
                 column   : byte;
                 param    : integer;
                 st       : str20;
               end;

  tCtrlRxRec = record
                 Msg     : tCtrlMsg;
                 curOp   : tCtrlMsg;
                 xValue  : integer;
                 yValue  : integer;
                 zValue  : integer;
               end;

  // MillAxes
  tStepPwr   = (pwrOff, pwrLow, pwrMed, pwrHigh);

  tAxCmd     = (acIdle, acDrivePwr, acJogSpeed, acTraverse, acStartSpeed,
                acMaxSpeed, acAccel, acClrAxesPos, acPanicStop, acStopMove,
                acClearTWI, acGetPos, acSetLimPol, acSetRefPol, acRefMove,
                acVelocity);

  tMoveMode  = (mvIdle, mvJog, mvTraverse, mvFeed, mvFeedXYZ, mvReference,
                mvVelocity);

  tAxeRxRec  = record
                 axCmd   : tAxCmd;
                 MoveMode: tMoveMode;
                 posAxe  : longint;
                 LimA    : boolean;
                 LimE    : boolean;
                 Home    : boolean;
               end;

  tAxeTxRecB = record
                 axCmd   : tAxCmd;
                 bb      : byte;
               end;

  tAxeTxRecW = record
                 axCmd   : tAxCmd;
                 ww      : word;
               end;

  tAxeTxRecI = record
                 axCmd   : tAxCmd;
                 ii      : integer;
               end;

  tAxeTxRecLi = record
                  axCmd   : tAxCmd;
                  Li      : longint;
                end;

  tAxeTxRecLw = record
                 axCmd    : tAxCmd;
                 Lw       : longword;
                end;

  tAxeRxRecLi = record
                  axCmd    : tAxCmd;
                  MoveMode : tMoveMode;
                  Li       : longint;
                end;


  // Host
  tMachStateE  = (tmDriveOn, tmSpindleOn, tmBusy, tmCOMfail, tmTimeOut,
                  tmParmError, tmOverTemp, tmInvSetup, tmLimSwX, tmLimSwY,
                  tmLimSwZ, tmRefSwX, tmRefSwY, tmRefSwZ, tmPanicStop);
  tMachState   = set of tMachStateE;

  tSetupStateE = (ssStartSpeed, ssAccel, ssMaxSpeed);
  tSetupState  = set of tSetupStateE;
  
  tAxeStateE   = (taXMove, taXJog, taYMove, taYJog, taZMove, taZJog, taLimSw, taRefSw);
  tAxeState    = set of tAxeStateE;

{--------------------------------------------------------------}
{ Const Declarations }
const
{$IFDEF LCD}
  msgNames  : array[tCtrlMsg] of string[10]
            = ('mIdle', 'mJogEna', 'mJogging', 'mSpindleOn', 'mStart', 'mWork',
               'mSpindleOff', 'mStop', 'mState', 'mBusy', 'mString', 'mParam');

  statNames : array[tMachStateE] of string[10]
            = ('tDriveOn', 'tSpindleOn', 'tBusy', 'tCOMfail', 'tTimeOut',
               'tParmError', 'tOverTemp', 'tInvSetup', 'tLimSwX', 'tLimSwY',
               'tLimSwZ', 'tRefSwX', 'tRefSwY', 'tRefSwZ', 'tPanicStop');
{$ENDIF}

  SOH       : char  = #1;
  STX       : char  = #2;
  ETX       : char  = #3;
  EOT       : char  = #4;
  ENQ       : char  = #5;
  ACK       : char  = #6;
  BEL       : char  = #7;
  BS        : char  = #8;
  HT        : char  = #9;
  LF        : char  = #10;
  VT        : char  = #11;
  FF        : char  = #12;
  CR        : char  = #13;
  SO        : char  = #14;
  SI        : char  = #15;
  DLE       : char  = #16;
  XON       : char  = #17;
  DC2       : char  = #18;
  XOFF      : char  = #19;
  DC4       : char  = #20;
  NAK       : char  = #21;
  SYN       : char  = #22;
  ETB       : char  = #23;
  CAN       : char  = #24;
  EM        : char  = #25;
  SUB       : char  = #26;
  ESC       : char  = #27;
  FS        : char  = #28;
  GS        : char  = #29;
  RS        : char  = #30;
  US        : char  = #31;

  MillSlvX  : byte = $10;                            // slave addr
  MillSlvY  : byte = $11;                            // slave addr
  MillSlvZ  : byte = $12;                            // slave addr
  MillCtrl  : byte = $03;                            // slave addr
  LM75      : byte = $49;                            // slave addr

  BufSize       : word = 1024;
  XONlim        : word = (BufSize * 7) div 10;       // 70%
  XOFFlim       : word = (BufSize * 8) div 10;       // 80%
  usecPerSecond : longword = 1000000;
  SlvTimeOut    : byte = 100;
  

{--------------------------------------------------------------}
{ Var Declarations }
{$IDATA}
var
  pAxeState[@PinB]      : byte;
  SX0[@pAxeState, 0]    : bit;
  SX1[@pAxeState, 1]    : bit;
  SY0[@pAxeState, 2]    : bit;
  SY1[@pAxeState, 3]    : bit;
  SZ0[@pAxeState, 4]    : bit;
  SZ1[@pAxeState, 6]    : bit;

  SwitchState[@PinC]    : byte;
  LimX[@SwitchState, 0] : bit;
  RefX[@SwitchState, 1] : bit;
  LimY[@SwitchState, 2] : bit;
  RefY[@SwitchState, 3] : bit;
  LimZ[@SwitchState, 4] : bit;
  RefZ[@SwitchState, 5] : bit;
  
  Coolant[@PortD, 6]    : bit;
  Spindle[@PortD, 7]    : bit;

  PanicSw[@PinE, 4]     : bit;
  StartSw[@PinE, 5]     : bit;

var
  AxeRxRec[@TWIRXBUFF]          : tAxeRxRec;
  AxeRxRecLi[@TWIRXBUFF]        : tAxeRxRecLi;
  CtrlRxRec[@TWIRXBUFF]         : tCtrlRxRec;
  rxMsg                         : tCtrlMsg;
  rxOp                          : tCtrlMsg;
  xv                            : integer;
  yv                            : integer;
  zv                            : integer;

  CtrlTxRec[@TWITXBUFF]         : tCtrlTxRec;
  AxeTxRecB[@TWITXBUFF]         : tAxeTxRecB;
  AxeTxRecW[@TWITXBUFF]         : tAxeTxRecW;
  AxeTxRecI[@TWITXBUFF]         : tAxeTxRecI;
  AxeTxRecLw[@TWITXBUFF]        : tAxeTxRecLw;
  AxeTxRecLi[@TWITXBUFF]        : tAxeTxRecLi;

  // timeout timer for TWInet Slaves
  NetTimeOut  : SysTimer8;

  // BitSet of TWIRxEmpty, TWIRxBusy, TWIRxFull, TWITxEmpty, TWITxBusy, TWITxFull, TWIbcCMD, TWIstatFail
  NetState    : tTWINetState;

  MachState   : tMachState;
  AxeState    : tAxeState;
  SetupState  : tSetupState;
  MoveMode    : tMoveMode;
  DrivePwr    : byte;
  StepPwr     : tStepPwr;
  MachTemp    : byte;
  SpindleRPM  : LongWord;

  StartSpeedX : word;
  StartSpeedY : word;
  StartSpeedZ : word;

  AccX        : word;
  AccY        : word;
  AccZ        : word;

  MaxSpeedX   : word;
  MaxSpeedY   : word;
  MaxSpeedZ   : word;

  FeedSpeedX  : longword;
  FeedSpeedY  : longword;
  FeedSpeedZ  : longword;
  
  AbsPosX     : longint;
  AbsPosY     : longint;
  AbsPosZ     : longint;

  JogSpeedX   : integer;
  JogSpeedY   : integer;
  JogSpeedZ   : integer;

  FastMoveX   : longint;
  FastMoveY   : longint;
  FastMoveZ   : longint;

  MoveX       : integer;
  MoveY       : integer;
  MoveZ       : integer;

  FeedStepsX  : longint;
  FeedStepsY  : longint;
  FeedStepsZ  : longint;
  StepBuffer  : array[0..BufSize-1] of byte;
  BuffInpPtr  : word;
  BuffOutPtr  : word;
  BuffCount   : word, locked;
  BuffXOFF    : boolean;
  FeedBusy    : boolean;
  FeedBkEnd   : boolean;
  stepXYZ     : byte;
  IntPtr      : pointer;
  
  TempTimer   : byte;
  
  Str10       : tStr10;
  bb, bb1     : byte;
  ww          : word;
  bool        : boolean;

{--------------------------------------------------------------}
{ functions }
interrupt Timer3CompA;
begin
  // only used for feed steps
  stepXYZ:= 0;
  if BuffCount > 0 then
    // fetch step info from buffer
    IntPtr:= @StepBuffer + BuffOutPtr;
    stepXYZ:= IntPtr^;
    dec(BuffCount);
    inc(BuffOutPtr);
    BuffOutPtr:= BuffOutPtr and (sizeOf(StepBuffer) -1);
    if (BuffCount = 0) and FeedBkEnd then
     FeedBusy:= false;
     FeedBkEnd:= false;
     TickTimerStop;
    endif;
  else
    if FeedStepsX > 0 then
      stepXYZ:= stepXYZ or $10;
      dec(FeedStepsX);
    elsif FeedStepsX < 0 then
      stepXYZ:= stepXYZ or $11;
      inc(FeedStepsX);
    endif;
    if FeedStepsY > 0 then
      stepXYZ:= stepXYZ or $20;
      dec(FeedStepsY);
    elsif FeedStepsY < 0 then
      stepXYZ:= stepXYZ or $22;
      inc(FeedStepsY);
    endif;
    if FeedStepsZ > 0 then
      stepXYZ:= stepXYZ or $40;
      dec(FeedStepsZ);
    elsif FeedStepsZ < 0 then
      stepXYZ:= stepXYZ or $44;
      inc(FeedStepsZ);
    endif;
    if stepXYZ = 0 then
      FeedBusy:= false;
      FeedBkEnd:= false;
      TickTimerStop;
    endif;
  endif;
  if stepXYZ <> 0 then
    // send steps to axe controllers
    PortA:= stepXYZ;
    if (stepXYZ and $11) = $10 then
      inc(AbsPosX);
    elsif (stepXYZ and $11) = $11 then
      dec(AbsPosX);
    endif;
    if (stepXYZ and $22) = $20 then
      inc(AbsPosY);
    elsif (stepXYZ and $22) = $22 then
      dec(AbsPosY);
    endif;
    if (stepXYZ and $44) = $40 then
      inc(AbsPosZ);
    elsif (stepXYZ and $44) = $44 then
      dec(AbsPosZ);
    endif;
    PortA:= stepXYZ and $0F;
  endif;
end;

procedure InitPorts;
begin
  PortA:= %00000000;
  DDRA:=  %11111111;
  PortB:= %11111111;
  DDRB:=  %00000000;
  PortC:= %11111111;
  DDRC:=  %00000000;
  PortD:= %00111111;
  DDRD:=  %11000000;
  PortE:= %11111100;
  DDRE:=  %00000000;
  PortF:= %11111111;
  DDRF:=  %00000000;
end InitPorts;

function RxHexByte : byte;
var
  st     : string[2];
  b      : byte;
begin
  Read(SerInp, st, 2);
  b:= HexToInt(st);
  return(b);
end;

function RxHexWord(lim : char) : word;
var
  st     : string[12];
  w      : word;
begin
  Read(SerInp, st, lim);
  w:= HexToInt(st);
  return(w);
end;

function RxHexInt(lim : char) : integer;
var
  st     : string[12];
  i      : integer;
begin
  Read(SerInp, st, lim);
  i:= HexToInt(st);
  return(i);
end;

function RxHexLw : longword;
var
  st     : string[12];
  lw     : longword;
begin
  Read(SerInp, st, '#');
  lw:= HexToInt(st);
  return(lw);
end;

function RxHexLI(lim : char) : longint;
var
  st     : string[12];
  li     : longint;
begin
  Read(SerInp, st, lim);
  li:= HexToInt(st);
  return(li);
end;


// Axe complete idle ??
function AxesIdle : boolean;
begin
  return((pAxeState and $3F) = $0);      // test SX0, SX1, SY0, SY1, SZ0, SZ1
end;

// Axe moves idle ??  Active Jog does not matter...
function AxesMoveIdle : boolean;
begin
  return((pAxeState and $15) = $0);      // test SX0, SY0, SZ0   = Move operation
end;

// Axe limit switches ??
function AxesSwitches : boolean;
begin
  return((SwitchState and $15) = $0);      // test SX2, SY2, SZ2   = limit switches
end;

Procedure SetSwitches;
var
  as        : tAxeState;
begin
  // taXMove, taXJog, taYMove, taYJog, taZMove, taZJog, taLimSw, taRefSw
  if FeedBusy or (BuffCount > 0) then
    as:= [taXMove, taYMove, taZMove];
  else
    as:= tAxeState(pAxeState and $3F);
  endif;

  if PanicSw then
    if not [tmPanicStop] in MachState then
      TickTimerStop;
      FeedSpeedX:= 0;
      FeedSpeedY:= 0;
      FeedSpeedZ:= 0;
      FeedStepsX:= 0;
      FeedStepsY:= 0;
      FeedStepsZ:= 0;
      FillBlock(@StepBuffer, sizeOf(StepBuffer), 0);
      BuffInpPtr:= 0;
      BuffOutPtr:= 0;
      BuffCount:= 0;
      FeedBusy:= false;
      MachState:= [tmInvSetup];
      SetupState:= [];
      // stop any movements
      TWItxBroadcast(byte(acPanicStop), 0, 1);                      // TWIcmd (slavenum), subnode, count
      Spindle:= 0;
      Coolant:= 0;
    endif;
    incl(MachState, [tmPanicStop]);
  else
    excl(MachState, [tmPanicStop]);
  endif;

  if DrivePwr <> 0 then
   incl(MachState, [tmDriveOn]);
  else
   excl(MachState, [tmDriveOn]);
  endif;
  if Spindle then
   incl(MachState, [tmSpindleOn]);
  else
   excl(MachState, [tmSpindleOn]);
  endif;
  if AxesIdle then
   excl(MachState, [tmBusy]);
  else
   incl(MachState, [tmBusy]);
  endif;
  if MachTemp > 50 then
   incl(MachState, [tmOverTemp]);
  else
   excl(MachState, [tmOverTemp]);
  endif;


  if LimX then
    incl(MachState, [tmLimSwX]);
    incl(as, [taLimSw]);
  else
    excl(MachState, [tmLimSwX]);
  endif;
  if LimY then
    incl(MachState, [tmLimSwY]);
    incl(as, [taLimSw]);
  else
    excl(MachState, [tmLimSwY]);
  endif;
  if LimZ then
    incl(MachState, [tmLimSwZ]);
    incl(as, [taLimSw]);
  else
    excl(MachState, [tmLimSwZ]);
  endif;

  if RefX then
    incl(MachState, [tmRefSwX]);
    incl(as, [taRefSw]);
  else
    excl(MachState, [tmRefSwX]);
  endif;
  if RefY then
    incl(MachState, [tmRefSwY]);
    incl(as, [taRefSw]);
  else
    excl(MachState, [tmRefSwY]);
  endif;
  if RefZ then
    incl(MachState, [tmRefSwZ]);
    incl(as, [taRefSw]);
  else
    excl(MachState, [tmRefSwZ]);
  endif;
  
  AxeState:= as;
end;

Procedure SetMachState(state : tMachState);
begin
  MachState:= [state];
  SetSwitches;
end;


function SendFrame(node, len : byte) : boolean;
var
  timeout  :  boolean;
begin
  SetSysTimer(NetTimeOut, SlvTimeOut);
  Timeout:= false;
  while not(TWItxFrame(node, len) or Timeout) do
    TWItxClear;
    mDelay(10);
    timeout:= isSysTimerZero(NetTimeOut);
  endwhile;
  if Timeout then
    TWItxClear;
    TWItxBroadcast(byte(acClearTWI), 0, 0);         // cmd, subnode, count
    mDelay(10);
    SetSysTimer(NetTimeOut, SlvTimeOut);
    Timeout:= false;
    while not(TWItxFrame(node, len) or Timeout) do
      TWItxClear;
      mDelay(10);
      timeout:= isSysTimerZero(NetTimeOut);
    endwhile;
  endif;
  if timeout then
    return(false);
  else
    return(true);
  endif;
end;

// Mill Controller = manual User interfasce
function MillCtrlTx(msg : tCtrlMsg; st : str20) : boolean;
begin
{$IFDEF LCD}
  LCDclrLine_M(LCD_m1, 3);
  Write(LCDOut_M, 'txMsg: '+msgNames[msg]+' ');
{$ENDIF}
{$IFDEF MillControl}
  CtrlTxRec.Msg:= msg;
  CtrlTxRec.st:= st;
  SetSysTimer(NetTimeOut, SlvTimeOut);
  if not SendFrame(MillCtrl, byte(sizeOf(tCtrlTxRec)));
    {$IFDEF LCD}
    LCDOut_M('?');
    {$ENDIF}
    return(false);
  else
    {$IFDEF LCD}
    LCDOut_M('!');
    {$ENDIF}
    return(true);
  endif;
{$ELSE}
  {$IFDEF LCD}
  LCDOut_M('!');
  {$ENDIF}
  return(true);
{$ENDIF}
end;

function MillCtrlRx(waitResponse : boolean) : boolean;
begin
  if waitResponse then
    {$IFDEF LCD}
    LCDclrLine_M(LCD_m1, 3);
    LCDclrLine_M(LCD_m1, 2);
    Write(LCDOut_M, 'rxMsg: ');
    {$ENDIF}
    SetSysTimer(NetTimeOut, 200);
    repeat
      mDelay(10);
    until TWIrxFrame(MillCtrl) or isSysTimerZero(NetTimeOut);
    if isSysTimerZero(NetTimeOut) then
      {$IFDEF LCD}
      Write(LCDOut_M, 'failed');
      {$ENDIF}
      TWIrxClear;
      return(false);
    else
      {$IFDEF LCD}
      rxMsg:= CtrlRxRec.Msg;
      rxOp:= CtrlRxRec.curOp;
      xv:= CtrlRxRec.xValue;
      yv:= CtrlRxRec.yValue;
      zv:= CtrlRxRec.zValue;
      Write(LCDOut_M, msgNames[rxMsg]+' !');
      {$ENDIF}
      TWIrxClear;
      return(true);
    endif;
  else
    if TWIrxFrame(MillCtrl) then
      rxMsg:= CtrlRxRec.Msg;
      rxOp:= CtrlRxRec.curOp;
      xv:= CtrlRxRec.xValue;
      yv:= CtrlRxRec.yValue;
      zv:= CtrlRxRec.zValue;
      {$IFDEF LCD}
      if rxMsg <> mdJogging then
        LCDclrLine_M(LCD_m1, 3);
        LCDclrLine_M(LCD_m1, 2);
        Write(LCDOut_M, 'rxMsg: '+ msgNames[rxMsg]);
      else
        LCDxy_M(LCD_m1, 0, 3);
        Write(LCDOut_M, IntToStr(xv:6)+IntToStr(yv:7)+IntToStr(zv:7));
      endif;
      {$ENDIF}
      TWIrxClear;
      return(true);
    else
      TWIrxClear;
      return(false);
    endif;
  endif;
end;



// For Traverse Operations
Procedure RecvStartSpeed;
begin
{$IFDEF LCD}
  LCDCLRLINE_M(LCD_m1, 2);
  Write(LCDOut_M, 'StartSpeed');
{$ENDIF}

//  result:= RecvStr(ESC+'S'+IntToHex(StartSpeedX, 2)+','
//                          +IntToHex(StartSpeedY, 2)+';'
//                          +IntToHex(StartSpeedZ, 2)+'.', true);
//  ACK/NAK

  StartSpeedX:= RxHexWord(',');
  StartSpeedY:= RxHexWord(';');
  StartSpeedZ:= RxHexWord('.');

  repeat
  until AxesIdle;

  AxeTxRecW.axCmd:= acStartSpeed;
  AxeTxRecW.ww:= StartSpeedX;
  if not SendFrame(MillSlvX, byte(sizeOf(AxeTxRecW))) then
    SerOut(NAK);
    return;
  endif;

  AxeTxRecW.ww:= StartSpeedY;
  if not SendFrame(MillSlvY, byte(sizeOf(AxeTxRecW))) then
    SerOut(NAK);
    return;
  endif;

  AxeTxRecW.ww:= StartSpeedZ;
  if not SendFrame(MillSlvZ, byte(sizeOf(AxeTxRecW))) then
    SerOut(NAK);
    return;
  endif;

  incl(SetupState, [ssStartSpeed]);
  if SetUpState = [ssStartSpeed, ssAccel, ssMaxSpeed] then
    excl(MachState, [tmInvSetup]);
  else
    incl(MachState, [tmInvSetup]);
  endif;
  SerOut(ACK);
end;

// For Traverse Operations
Procedure RecvAcceleration;
begin
{$IFDEF LCD}
  LCDCLRLINE_M(LCD_m1, 2);
  Write(LCDOut_M, 'Acceleration');
{$ENDIF}

//  result:= RecvStr(ESC+'A'+IntToHex(AccX, 2)+','
//                          +IntToHex(AccY, 2)+';'
//                          +IntToHex(AccZ, 2)+'.', true);
//  ACK/NAK

  AccX:= RxHexWord(',');
  AccY:= RxHexWord(';');
  AccZ:= RxHexWord('.');

  repeat
  until AxesIdle;

  AxeTxRecW.axCmd:= acAccel;
  AxeTxRecW.ww:= AccX;
  if not SendFrame(MillSlvX, byte(sizeOf(AxeTxRecW))) then
    SerOut(NAK);
    return;
  endif;

  AxeTxRecW.ww:= AccY;
  if not SendFrame(MillSlvY, byte(sizeOf(AxeTxRecW))) then
    SerOut(NAK);
    return;
  endif;

  AxeTxRecW.ww:= AccZ;
  if not SendFrame(MillSlvZ, byte(sizeOf(AxeTxRecW))) then
    SerOut(NAK);
    return;
  endif;

  incl(SetupState, [ssAccel]);
  if SetUpState = [ssStartSpeed, ssAccel, ssMaxSpeed] then
    excl(MachState, [tmInvSetup]);
  else
    incl(MachState, [tmInvSetup]);
  endif;
  SerOut(ACK);
end;

// For Traverse Operations
Procedure RecvMaxSpeed;
begin
{$IFDEF LCD}
  LCDCLRLINE_M(LCD_m1, 2);
  Write(LCDOut_M, 'MaxSpeed');
{$ENDIF}

//  result:= RecvStr(ESC+'M'+IntToHex(MaxSpeedX, 2)+','
//                          +IntToHex(MaxSpeedY, 2)+';'
//                          +IntToHex(MaxSpeedZ, 2)+'.', true);
//  ACK/NAK

  MaxSpeedX:= RxHexWord(',');
  MaxSpeedY:= RxHexWord(';');
  MaxSpeedZ:= RxHexWord('.');

  repeat
  until AxesIdle;

  AxeTxRecW.axCmd:= acMaxSpeed;
  AxeTxRecW.ww:= MaxSpeedX;
  if not SendFrame(MillSlvX, byte(sizeOf(AxeTxRecW))) then
    SerOut(NAK);
    return;
  endif;

  AxeTxRecW.ww:= MaxSpeedY;
  if not SendFrame(MillSlvY, byte(sizeOf(AxeTxRecW))) then
    SerOut(NAK);
    return;
  endif;

  AxeTxRecW.ww:= MaxSpeedZ;
  if not SendFrame(MillSlvZ, byte(sizeOf(AxeTxRecW))) then
    SerOut(NAK);
    return;
  endif;

  incl(SetupState, [ssMaxSpeed]);
  if SetUpState = [ssStartSpeed, ssAccel, ssMaxSpeed] then
    excl(MachState, [tmInvSetup]);
  else
    incl(MachState, [tmInvSetup]);
  endif;
  SerOut(ACK);
end;

// For Feed/Work Operations
Procedure RecvFeedSpeed;
var
  speed   : word;
begin
{$IFDEF LCD}
  LCDCLRLINE_M(LCD_m1, 2);
  Write(LCDOut_M, 'FeedSpeed');
{$ENDIF}

//  result:= RecvStr(FS+IntToHex(FeedSpeedX, 2)+','
//                     +IntToHex(FeedSpeedY, 2)+';'
//                     +IntToHex(FeedSpeedZ, 2)+'.', true);
//  result = steps/sec
//  ACK/NAK

  speed:= RxHexWord(',');
  FeedSpeedX:= usecPerSecond div longword(speed);
  speed:= RxHexWord(';');
  FeedSpeedY:= usecPerSecond div longword(speed);
  speed:= RxHexWord('.');
  FeedSpeedZ:= usecPerSecond div longword(speed);
  SerOut(ACK);
end;



// ==========================================
// Housekeeping commands
// ==========================================
Procedure RequestMachineState;
begin
{$IFDEF LCD}
  LCDCLRLINE_M(LCD_m1, 2);
  Write(LCDOut_M, 'MachineState');
{$ENDIF}

//  COMport.TransmitChar(ENQ);
//  <ENQ><nn>#

  SerOut(ENQ);
  Str10:= IntToHex(word(MachState));
  Write(SerOut, Str10);
  SerOut('#');
end;

Procedure RequestAxeState;
var
  b       : byte;
begin
{$IFDEF LCD}
  LCDCLRLINE_M(LCD_m1, 2);
  Write(LCDOut_M, 'AxeState');
{$ENDIF}

//  COMport.TransmitChar(HT);
//  <HT><nn>#

  SerOut(HT);
  Str10:= ByteToHex(AxeState);
  Write(SerOut, Str10);
  SerOut('#');
end;

procedure RequestStart;
begin
  if INP_RAISE1(5) then
    // Limit or Panic Switch active ??
    if ([taLimSw] in AxeState) or ([tmPanicStop] in MachState) then
      SerOut(NAK);
    else
      SerOut(ACK);
    endif;
  else
    SerOut(NAK);
  endif;
end;

Procedure RecvSync;
begin
//  COMport.TransmitChar(SYN);
end;

procedure RecvLimSwitchPol;
var
  temp    : byte;
  ok      : boolean;
  ch      : char;
begin
  ok:= true;
{$IFDEF LCD}
  LCDCLRLINE_M(LCD_m1, 2);
  Write(LCDOut_M, 'Lim Switch Polarity');
{$ENDIF}

//  result:= RecvStr(<SO><px>,<py>;<pz>.);
//  ACK/NAK

  temp:= RxHexByte;
  ch:= SerInp;                           // get delimiter
  AxeTxRecB.axCmd:= acSetLimPol;
  AxeTxRecB.bb:= temp;
  if not SendFrame(MillSlvX, byte(sizeOf(AxeTxRecB))) then
    ok:= false;
  endif;

  temp:= RxHexByte;
  ch:= SerInp;                           // get delimiter
  AxeTxRecB.axCmd:= acSetLimPol;
  AxeTxRecB.bb:= temp;
  if not SendFrame(MillSlvY, byte(sizeOf(AxeTxRecB))) then
    ok:= false;
  endif;

  temp:= RxHexByte;
  ch:= SerInp;                           // get delimiter
  AxeTxRecB.axCmd:= acSetLimPol;
  AxeTxRecB.bb:= temp;
  if not SendFrame(MillSlvZ, byte(sizeOf(AxeTxRecB))) then
    ok:= false;
  endif;

  if ok then
    SerOut(ACK);
  else
    SerOut(NAK);
  endif;
end;

procedure RecvRefSwitchPol;
var
  temp    : byte;
  ok      : boolean;
  ch      : char;
begin
  ok:= true;
{$IFDEF LCD}
  LCDCLRLINE_M(LCD_m1, 2);
  Write(LCDOut_M, 'Ref Switch Polarity');
{$ENDIF}

//  result:= RecvStr(<DLE><px>,<py>;<pz>.);
//  ACK/NAK

  temp:= RxHexByte;
  ch:= SerInp;                           // get delimiter
  AxeTxRecB.axCmd:= acSetRefPol;
  AxeTxRecB.bb:= temp;
  if not SendFrame(MillSlvX, byte(sizeOf(AxeTxRecB))) then
    ok:= false;
  endif;

  temp:= RxHexByte;
  ch:= SerInp;                           // get delimiter
  AxeTxRecB.axCmd:= acSetRefPol;
  AxeTxRecB.bb:= temp;
  if not SendFrame(MillSlvY, byte(sizeOf(AxeTxRecB))) then
    ok:= false;
  endif;

  temp:= RxHexByte;
  ch:= SerInp;                           // get delimiter
  AxeTxRecB.axCmd:= acSetRefPol;
  AxeTxRecB.bb:= temp;
  if not SendFrame(MillSlvZ, byte(sizeOf(AxeTxRecB))) then
    ok:= false;
  endif;

  if ok then
    SerOut(ACK);
  else
    SerOut(NAK);
  endif;
end;

Procedure RequestTemperature;
begin
{$IFDEF LCD}
  LCDCLRLINE_M(LCD_m1, 2);
  Write(LCDOut_M, 'Temperature');
{$ENDIF}

//  COMport.TransmitChar(ETB);
//  <ETB><bb>#

  SerOut(ETB);
  Str10:= ByteToHex(MachTemp);
  Write(SerOut, Str10);
  SerOut('#');
end;

Procedure RequestAbsPosition;
begin
{$IFDEF LCD}
  LCDCLRLINE_M(LCD_m1, 2);
  Write(LCDOut_M, 'AbsPosition');
{$ENDIF}

//  COMport.TransmitChar(ETX);
//  <ETX><xx>,<yy>;<zz>.

  if MoveMode = mvTraverse then
    // request current slave steps or position
    TWItxBroadcast(byte(acGetPos), 0, 0);                  // TWIcmd (slavenum), subnode, count

    // request x position
    SetSysTimer(NetTimeOut, 200);
    repeat
      mDelay(20);
    until TWIrxFrame(MillSlvX) or isSysTimerZero(NetTimeOut);
    if isSysTimerZero(NetTimeOut) then
      AbsPosX:= 0;
    elsif Sign(AbsPosX) then
      AbsPosX:= AbsPosX - AxeRxRecLi.Li;
    else
      AbsPosX:= AbsPosX + AxeRxRecLi.Li;
    endif;

    // request y position
    SetSysTimer(NetTimeOut, 200);
    repeat
      mDelay(20);
    until TWIrxFrame(MillSlvY) or isSysTimerZero(NetTimeOut);
    if isSysTimerZero(NetTimeOut) then
      AbsPosY:= 0;
    elsif Sign(AbsPosY) then
      AbsPosY:= AbsPosY - AxeRxRecLi.Li;
    else
      AbsPosY:= AbsPosY + AxeRxRecLi.Li;
    endif;

    // request z position
    SetSysTimer(NetTimeOut, 200);
    repeat
      mDelay(20);
    until TWIrxFrame(MillSlvZ) or isSysTimerZero(NetTimeOut);
    if isSysTimerZero(NetTimeOut) then
      AbsPosZ:= 0;
    elsif Sign(AbsPosZ) then
      AbsPosZ:= AbsPosZ - AxeRxRecLi.Li;
    else
      AbsPosZ:= AbsPosZ + AxeRxRecLi.Li;
    endif;
  elsif MoveMode in[mvIdle, mvJog] then      // no position detection possible
    AbsPosX:= 0;
    AbsPosY:= 0;
    AbsPosZ:= 0;
  endif;

  SerOut(ETX);
  Write(SerOut, LongToHex(AbsPosX)+',');
  Write(SerOut, LongToHex(AbsPosY)+';');
  Write(SerOut, LongToHex(AbsPosZ)+'.');
end;


// ==========================================
// Immediate commands
// ==========================================
Procedure RecvPanicStop;
begin
{$IFDEF LCD}
  LCDCLRLINE_M(LCD_m1, 2);
  Write(LCDOut_M, 'PanicStop');
{$ENDIF}

//  result:= RecvCMD(BEL, 1000);
//  ACK/NAK

  TickTimerStop;
  FeedSpeedX:= 0;
  FeedSpeedY:= 0;
  FeedSpeedZ:= 0;
  FeedStepsX:= 0;
  FeedStepsY:= 0;
  FeedStepsZ:= 0;
  FillBlock(@StepBuffer, sizeOf(StepBuffer), 0);
  BuffInpPtr:= 0;
  BuffOutPtr:= 0;
  BuffCount:= 0;
  FeedBusy:= false;
  SetMachState([tmInvSetup]);
  SetupState:= [];
  // stop any movements
  TWItxBroadcast(byte(acPanicStop), 0, 1);                      // TWIcmd (slavenum), subnode, count
  Spindle:= 0;
  Coolant:= 0;
  SerOut(ACK);
end;

Procedure RecvStopMove;
begin
{$IFDEF LCD}
  LCDCLRLINE_M(LCD_m1, 2);
  Write(LCDOut_M, 'StopMove');
{$ENDIF}

//  result:= RecvCMD(XOFF, 2000);
//  ACK/NAK

  TWItxBroadcast(byte(acStopMove), 0, 1);                      // TWIcmd (slavenum), subnode, count
  // wenn FeedStepsNN luft, hier Position rechnen
  // ansonsten von der Achse abfragen
  // ??????????????????????????????????????????
  SerOut(ACK);
end;

Procedure RecvFlushBuffer;
begin
{$IFDEF LCD}
  LCDCLRLINE_M(LCD_m1, 2);
  Write(LCDOut_M, 'FlushBuffer');
{$ENDIF}

//  result:= RecvCMD(GS, 2000);
//  ACK/NAK

  BuffInpPtr:= 0;
  BuffOutPtr:= 0;
  BuffCount:= 0;
  SerOut(ACK);
end;

Procedure RecvDrivePWR;
var
  ok        : boolean;
begin
//  result:= RecvCMD(STX+IntToHex(DrivePWR, 2), 2000);
//  ACK/NAK

  DrivePwr:= RxHexByte;
  
  if [tmPanicStop] in MachState then
    SerOut(NAK);
    return;
  endif;

  if DrivePwr < 25 then
    StepPwr:= pwrOff;
  elsif DrivePwr < 50 then
    StepPwr:= pwrLow;
  elsif DrivePwr < 75 then
    StepPwr:= pwrMed;
  else
    StepPwr:= pwrHigh;
  endif;
  
{$IFDEF LCD}
  LCDCLRLINE_M(LCD_m1, 3);
  Write(LCDOut_M, 'DrivePwr '+ByteToStr(DrivePwr));
{$ENDIF}
  ok:= true;

  AxeTxRecB.axCmd:= acDrivePwr;
  AxeTxRecB.bb:= byte(StepPwr);
  if not SendFrame(MillSlvX, byte(sizeOf(AxeTxRecB))) then
    ok:= false;
   {$IFDEF LCD}
    Write(LCDOut_M, ' X');
   {$ENDIF}
  endif;

  if not SendFrame(MillSlvY, byte(sizeOf(AxeTxRecB))) then
    ok:= false;
   {$IFDEF LCD}
    Write(LCDOut_M, ' Y');
   {$ENDIF}
  endif;

  if not SendFrame(MillSlvZ, byte(sizeOf(AxeTxRecB))) then
    ok:= false;
   {$IFDEF LCD}
    Write(LCDOut_M, ' Z');
   {$ENDIF}
  endif;

  if ok then
    SerOut(ACK);
  else
    SerOut(NAK);
  endif;
end;

Procedure RecvSpindleRPM;
begin
{$IFDEF LCD}
  LCDCLRLINE_M(LCD_m1, 2);
  Write(LCDOut_M, 'SpindleRPM');
{$ENDIF}

//  result:= RecvCMD(SOH+IntToHex(SpindleRPM, 4)+'#', 2000);
//  ACK/NAK

  SpindleRPM:= RxHexLw;

  if [tmPanicStop] in MachState then
    SerOut(NAK);
    return;
  endif;

  if SpindleRPM = 0 then
    Spindle:= 0;
  else
    Spindle:= 1;
  endif;
  SerOut(ACK);
end;

Procedure RecvCoolingOnOff;
var
  b         : byte;
begin
{$IFDEF LCD}
  LCDCLRLINE_M(LCD_m1, 2);
  Write(LCDOut_M, 'Cooling');
{$ENDIF}

//  result:= RecvCMD(SI+'FF', 2000);

//  ACK/NAK

  b:= RxHexByte;
  
  if [tmPanicStop] in MachState then
    SerOut(NAK);
    return;
  endif;

  if b = $FF then
    Coolant:= 1;
  else
    Coolant:= 0;
  endif;
  SerOut(ACK);
end;


// =======================================
// Move commands
// =======================================

// Jog Axes. Only the desired speed and direction
// is passed in integer values
Procedure RecvJogSpeed;
var
  ok    : boolean;
begin
{$IFDEF LCD}
  LCDCLRLINE_M(LCD_m1, 2);
  Write(LCDOut_M, 'JogSpeed');
{$ENDIF}

//  result:= RecvStr(<DC4><xx>,<yy>;<zz>.);
//  ACK/NAK

  JogSpeedX:= RxHexInt(',');
  JogSpeedY:= RxHexInt(';');
  JogSpeedZ:= RxHexInt('.');
  
  // Panic Switch active ??
  if [tmPanicStop] in MachState then
    SerOut(NAK);
    return;
  endif;

  MoveMode:= mvJog;

  repeat
  until AxesMoveIdle;
  
  ok:= true;

  AxeTxRecI.axCmd:= acJogSpeed;
  if JogSpeedX <> 0 then
    AxeTxRecI.ii:= 20000 div JogSpeedX;
  else
    AxeTxRecI.ii:= 0;
  endif;
  if not SendFrame(MillSlvX, byte(sizeOf(AxeTxRecI))) then
    ok:= false;
   {$IFDEF LCD}
    Write(LCDOut_M, ' X');
   {$ENDIF}
  endif;

  if JogSpeedY <> 0 then
    AxeTxRecI.ii:= 20000 div JogSpeedY;
  else
    AxeTxRecI.ii:= 0;
  endif;
  if not SendFrame(MillSlvY, byte(sizeOf(AxeTxRecI))) then
    ok:= false;
   {$IFDEF LCD}
    Write(LCDOut_M, ' Y');
   {$ENDIF}
  endif;

  if JogSpeedZ <> 0 then
    AxeTxRecI.ii:= 20000 div JogSpeedZ;
  else
    AxeTxRecI.ii:= 0;
  endif;
  if not SendFrame(MillSlvZ, byte(sizeOf(AxeTxRecI))) then
    ok:= false;
   {$IFDEF LCD}
    Write(LCDOut_M, ' Z');
   {$ENDIF}
  endif;

  if ok then
    SerOut(ACK);
  else
    SerOut(NAK);
  endif;
end;

procedure RecvVelocityXYZ;
var
  ok     : boolean;
begin
{$IFDEF LCD}
  LCDCLRLINE_M(LCD_m1, 2);
  Write(LCDOut_M, 'VelocityXYZ');
{$ENDIF}

//  result:= SendStr(CAN+Xstr+Ystr+Zstr, true);
//  ACK/NAK

  MoveX:= RxHexInt(',');
  MoveY:= RxHexInt(';');
  MoveZ:= RxHexInt('.');

  // Limit or Panic Switch active ??
  if ([taLimSw] in AxeState) or ([tmPanicStop] in MachState) then
    SerOut(NAK);
    return;
  endif;


{$IFDEF LCD}
  Write(LCDOut_M, ' '+IntToStr(MoveX)+' '+IntToStr(MoveY)+' '+IntToStr(MoveZ));
{$ENDIF}

  if not MoveMode in[mvVelocity, mvIdle] then
    repeat
    until AxesIdle;
  endif;
  MoveMode:= mvVelocity;

  ok:= true;
  AxeTxRecI.axCmd:= acVelocity;
  AxeTxRecI.ii:= MoveX;
  if not SendFrame(MillSlvX, byte(sizeOf(AxeTxRecI))) then
    ok:= false;
  endif;

  AxeTxRecI.axCmd:= acVelocity;
  AxeTxRecI.ii:= MoveY;
  if not SendFrame(MillSlvY, byte(sizeOf(AxeTxRecI))) then
    ok:= false;
  endif;

  AxeTxRecI.axCmd:= acVelocity;
  AxeTxRecI.ii:= MoveZ;
  if not SendFrame(MillSlvZ, byte(sizeOf(AxeTxRecI))) then
    ok:= false;
  endif;

  if ok then
    SerOut(ACK);
  else
    SerOut(NAK);
  endif;
end;

Procedure RecvRefMoveXYZ;
var
  ok     : boolean;
begin
{$IFDEF LCD}
  LCDCLRLINE_M(LCD_m1, 2);
  Write(LCDOut_M, 'RefMoveXYZ');
{$ENDIF}

//  result:= SendStr(US+Xstr+Ystr+Zstr, true);
//  ACK/NAK

  MoveX:= RxHexInt(',');
  MoveY:= RxHexInt(';');
  MoveZ:= RxHexInt('.');
  
  // Limit or Panic Switch active ??
  if ([taLimSw] in AxeState) or ([tmPanicStop] in MachState) then
    SerOut(NAK);
    return;
  endif;
  
  MoveMode:= mvReference;

{$IFDEF LCD}
  Write(LCDOut_M, ' '+IntToStr(MoveX)+' '+IntToStr(MoveY)+' '+IntToStr(MoveZ));
{$ENDIF}

  repeat
  until AxesIdle;

  ok:= true;
  if MoveX <> 0 then
    AxeTxRecI.axCmd:= acRefMove;
    AxeTxRecI.ii:= MoveX;
    if not SendFrame(MillSlvX, byte(sizeOf(AxeTxRecI))) then
      ok:= false;
    endif;
  endif;

  if MoveY <> 0 then
    AxeTxRecI.axCmd:= acRefMove;
    AxeTxRecI.ii:= MoveY;
    if not SendFrame(MillSlvY, byte(sizeOf(AxeTxRecI))) then
      ok:= false;
    endif;
  endif;

  if MoveZ <> 0 then
    AxeTxRecI.axCmd:= acRefMove;
    AxeTxRecI.ii:= MoveZ;
    if not SendFrame(MillSlvZ, byte(sizeOf(AxeTxRecI))) then
      ok:= false;
    endif;
  endif;

  if ok then
    SerOut(ACK);
  else
    SerOut(NAK);
  endif;
end;

Procedure RecvFastMoveXYZ;
var
  ok     : boolean;
begin
{$IFDEF LCD}
  LCDCLRLINE_M(LCD_m1, 2);
  Write(LCDOut_M, 'MoveXYZ');
{$ENDIF}

//  result:= SendStr(EOT+Xstr+Ystr+Zstr, true);
//  ACK/NAK

  FastMoveX:= RxHexLI(',');
  FastMoveY:= RxHexLI(';');
  FastMoveZ:= RxHexLI('.');

  // Limit or Panic Switch active ??
  if ([taLimSw] in AxeState) or ([tmPanicStop] in MachState) then
    SerOut(NAK);
    return;
  endif;

  MoveMode:= mvTraverse;

{$IFDEF LCD}
  Write(LCDOut_M, ' '+LongToStr(FastMoveX)+' '+LongToStr(FastMoveY)+' '+LongToStr(FastMoveZ));
{$ENDIF}

  repeat
  until AxesIdle;

  ok:= true;
  if FastMoveX <> 0 then
    AxeTxRecLi.axCmd:= acTraverse;
    AxeTxRecLi.Li:= FastMoveX;
    if not SendFrame(MillSlvX, byte(sizeOf(AxeTxRecLi))) then
      ok:= false;
    endif;
  endif;

  if FastMoveY <> 0 then
    AxeTxRecLi.axCmd:= acTraverse;
    AxeTxRecLi.Li:= FastMoveY;
    if not SendFrame(MillSlvY, byte(sizeOf(AxeTxRecLi))) then
      ok:= false;
    endif;
  endif;

  if FastMoveZ <> 0 then
    AxeTxRecLi.axCmd:= acTraverse;
    AxeTxRecLi.Li:= FastMoveZ;
    if not SendFrame(MillSlvZ, byte(sizeOf(AxeTxRecLi))) then
      ok:= false;
    endif;
  endif;

  if ok then
    SerOut(ACK);
  else
    SerOut(NAK);
  endif;
end;

Procedure RecvFastMoveXY;
var
  ok     : boolean;
begin
{$IFDEF LCD}
  LCDCLRLINE_M(LCD_m1, 2);
  Write(LCDOut_M, 'MoveXY');
{$ENDIF}

//  result:= RecvStr(FF+Xstr+Ystr, true);
//  ACK/NAK

  FastMoveX:= RxHexLI(',');
  FastMoveY:= RxHexLI(';');

  // Limit or Panic Switch active ??
  if ([taLimSw] in AxeState) or ([tmPanicStop] in MachState) then
    SerOut(NAK);
    return;
  endif;

  MoveMode:= mvTraverse;

{$IFDEF LCD}
  Write(LCDOut_M, ' '+LongToStr(FastMoveX)+' '+LongToStr(FastMoveY));
{$ENDIF}

  repeat
  until AxesIdle;

  ok:= true;
  if FastMoveX <> 0 then
    AxeTxRecLi.axCmd:= acTraverse;
    AxeTxRecLi.Li:= FastMoveX;
    if not SendFrame(MillSlvX, byte(sizeOf(AxeTxRecLi))) then
      ok:= false;
    endif;
  endif;

  if FastMoveY <> 0 then
    AxeTxRecLi.axCmd:= acTraverse;
    AxeTxRecLi.Li:= FastMoveY;
    if not SendFrame(MillSlvY, byte(sizeOf(AxeTxRecLi))) then
      ok:= false;
    endif;
  endif;
  if ok then
    SerOut(ACK);
  else
    SerOut(NAK);
  endif;
end;

Procedure RecvFastMoveZ;
begin
{$IFDEF LCD}
  LCDCLRLINE_M(LCD_m1, 2);
  Write(LCDOut_M, 'MoveZ');
{$ENDIF}

//  result:= RecvStr(BS+Zstr, true);
//  ACK/NAK

  FastMoveZ:= RxHexLI('.');

  // Limit or Panic Switch active ??
  if ([taLimSw] in AxeState) or ([tmPanicStop] in MachState) then
    SerOut(NAK);
    return;
  endif;

  MoveMode:= mvTraverse;

  repeat
  until AxesIdle;

  if FastMoveZ <> 0 then
    AxeTxRecLi.axCmd:= acTraverse;
    AxeTxRecLi.Li:= FastMoveZ;
    if not SendFrame(MillSlvZ, byte(sizeOf(AxeTxRecLi))) then
      SerOut(NAK);
      return;
    endif;
  endif;
  SerOut(ACK);
{$IFDEF LCD}
  Write(LCDOut_M, ' '+LongToStr(FastMoveZ));
{$ENDIF}
end;

Procedure RecvEndOfFeed;
begin
{$IFDEF LCD}
  LCDCLRLINE_M(LCD_m1, 2);
  Write(LCDOut_M, 'EndOfFeed');
{$ENDIF}

  // the master sends this request after a block of feed steps
  // it notices the slave that the block transfer is finished now

//  result:= RecvCMD(CR, 2000);
//  ACK/NAK

  FeedBkEnd:= true;
  
  // Limit or Panic Switch active ??
  if ([taLimSw] in AxeState) or ([tmPanicStop] in MachState) then
    SerOut(NAK);
  else
    SerOut(ACK);
  endif;
end;

// each byte cmd contains 1 step and 1 dir info for each axe
// bit0  = direction X
// bit4  = do step X
// bit1  = direction Y
// bit5  = do step Y
// bit2  = direction Y
// bit6  = do step Z
// all other bits reserved = 0
// All cmds are placed into the cmd buffer
// The TickTimer int fetches these cmds with feedrate intervals
// and writes it via port bits to the axes slaves
Procedure RecvFeedStepsXYZ;
var
  BuffPtr    : pointer;
  steps      : byte;
begin
(*
{$IFDEF LCD}
  LCDCLRLINE_M(LCD_m1, 2);
  Write(LCDOut_M, 'FeedStepsXYZ');
{$ENDIF}
*)
//  RecvStr(DC2+IntToHex(FeedSteps, 2), false);
//  XON/XOFF

  steps:= RxHexByte;

  // Limit or Panic Switch active ??
  if ([taLimSw] in AxeState) or ([tmPanicStop] in MachState) then
    SerOut(XOFF);
    return;
  endif;

  // store step info into buffer
  BuffPtr:= @StepBuffer + BuffInpPtr;
  BuffPtr^:= steps;
  inc(BuffCount);
  inc(BuffInpPtr);
  BuffInpPtr:= BuffInpPtr and (sizeOf(StepBuffer) -1);
  MoveMode:= mvFeedXYZ;

  if BuffCount > XOFFlim then
    SerOut(XOFF);
    BuffXOFF:= true;
  endif;
end;

procedure RecvClearPosition;
begin
{$IFDEF LCD}
  LCDCLRLINE_M(LCD_m1, 2);
  Write(LCDOut_M, 'ClearPosition');
{$ENDIF}

//  result:= RecvStr(RS, true);
//  ACK/NAK

  // Limit or Panic Switch active ??
  if ([taLimSw] in AxeState) or ([tmPanicStop] in MachState) then
    SerOut(NAK);
    return;
  endif;

  repeat
  until AxesIdle;

  TWItxBroadcast(byte(acClrAxesPos), 0, 0);                  // TWIcmd (slavenum), subnode, count
  AbsPosX:= 0;
  AbsPosY:= 0;
  AbsPosZ:= 0;
  SerOut(ACK);
end;

Procedure RecvFeedStepsX;
begin
{$IFDEF LCD}
  LCDCLRLINE_M(LCD_m1, 2);
  Write(LCDOut_M, 'FeedStepsX');
{$ENDIF}

//  result:= RecvStr(EM+Xstr, true);
//  ACK/NAK

  FeedStepsX:= RxHexLI(',');
  FeedStepsY:= 0;
  FeedStepsZ:= 0;

  // Limit or Panic Switch active ??
  if ([taLimSw] in AxeState) or ([tmPanicStop] in MachState) then
    SerOut(NAK);
    return;
  endif;

  MoveMode:= mvFeed;
  AbsPosX:= 0;
  AbsPosY:= 0;
  AbsPosZ:= 0;
  SerOut(ACK);
end;

Procedure RecvFeedStepsY;
begin
{$IFDEF LCD}
  LCDCLRLINE_M(LCD_m1, 2);
  Write(LCDOut_M, 'FeedStepsY');
{$ENDIF}

//  result:= RecvStr(SUB+Ystr, true);
//  ACK/NAK

  FeedStepsX:= 0;
  FeedStepsY:= RxHexLI(';');
  FeedStepsZ:= 0;

  // Limit or Panic Switch active ??
  if ([taLimSw] in AxeState) or ([tmPanicStop] in MachState) then
    SerOut(NAK);
    return;
  endif;

  MoveMode:= mvFeed;
  AbsPosX:= 0;
  AbsPosY:= 0;
  AbsPosZ:= 0;
  SerOut(ACK);
end;

Procedure RecvFeedStepsZ;
begin
{$IFDEF LCD}
  LCDCLRLINE_M(LCD_m1, 2);
  Write(LCDOut_M, 'FeedStepsZ');
{$ENDIF}

//  result:= RecvStr(VT+Zstr, true);
//  ACK/NAK

  FeedStepsX:= 0;
  FeedStepsY:= 0;
  FeedStepsZ:= RxHexLI('.');

  // Limit or Panic Switch active ??
  if ([taLimSw] in AxeState) or ([tmPanicStop] in MachState) then
    SerOut(NAK);
    return;
  endif;

  MoveMode:= mvFeed;
  AbsPosX:= 0;
  AbsPosY:= 0;
  AbsPosZ:= 0;
  SerOut(ACK);
end;

procedure HandleStopp;
begin
   if not MillCtrlTx(mdStop, '') then
     // error handling
   endif;
   mDelay(1000);
   // MillController must be "idle" here
  repeat
    if not MillCtrlRx(true) then
      // timeout error handling
    elsif rxMsg <> mdIdle then
      {$IFDEF LCD}
      LCDCLR_M(LCD_m1);
      Write(LCDOut_M, 'Idle mode expected');
      mDelay(2000);
      {$ENDIF}
      if not MillCtrlTx(mdState, '') then
        // error handling
      elsif not MillCtrlRx(true) then
      // timeout error handling
      endif;
    endif;
  until rxMsg = mdIdle;
end;

procedure HandleJogging;
begin
  repeat
    repeat
      mDelay(10);
    until MillCtrlRx(false);
    if rxMsg = mdJogging then
      // process jog values
      // .........
      // xv
      // yv
      // zv
    elsif rxMsg = mdJog then
      if not MillCtrlTx(mdJog, '') then
        // error handling
      endif;
      // stop jogging
      // .........
    else
      // stop jogging
      // .........
    endif;
  until rxMsg <> mdJogging;
end;

procedure TestMillControlDevice;
begin
  loop
    repeat
      // MillController must be "idle" here
      if not MillCtrlTx(mdState, '') then
        // error handling
      endif;
      if not MillCtrlRx(true) then
        // timeout error handling
      elsif rxMsg <> mdIdle then
        {$IFDEF LCD}
        LCDCLR_M(LCD_m1);
        Write(LCDOut_M, 'Idle mode expected');
        mDelay(2000);
        {$ENDIF}
        if not MillCtrlTx(mdIdle, '') then
          // error handling
        elsif not MillCtrlRx(true) then
        // timeout error handling
        endif;
      endif;
    until rxMsg = mdIdle;

    // we expect a "jog" response
    repeat
      repeat
        mDelay(10);
      until MillCtrlRx(false);
      if rxMsg = mdJog then
        if not MillCtrlTx(mdJog, '') then
          // error handling
        endif;
      elsif rxMsg <> mdIdle then
        // message is <> jog, maybe an emergency stopp
        if rxMsg = mdStop then
          // handle stopp condition
          HandleStopp;
          return;
        else
          // error handling
        endif;
      endif;
    until rxMsg <> mdIdle;

    // we expect a "jogging" or a "start" response
    repeat
      repeat
        mDelay(10);
      until MillCtrlRx(false);
      if rxMsg = mdStart then
        if not MillCtrlTx(mdStart, '') then
          // error handling
        endif;
      elsif rxMsg = mdJogging then
        if MillCtrlTx(mdJogging, '') then
          HandleJogging;
          if rxMsg = mdStop then
            // handle stopp condition
            HandleStopp;
            return;
          endif;
        else
          // error handling
        endif;
      else
        // message is <> jogging/start, maybe an emergency stopp
        if rxMsg = mdStop then
          // handle stopp condition
          HandleStopp;
          return;
        else
          // error handling
        endif;
        continue;
      endif;
    until rxMsg <> mdJog;
    
    // we do the whole work
    if not MillCtrlTx(mdStart, '') then
      // error handling
    endif;

    repeat
      // wating for the start button
      repeat
        mDelay(10);
      until MillCtrlRx(false);
      if rxMsg = mdStart then
        Spindle:= 1;
      else
        // message is <> mdStart, maybe an emergency stopp
        if rxMsg = mdStop then
          // handle stopp condition
          HandleStopp;
          return;
        else
          // error handling
        endif;
      endif;

      // wating for spindle run
      mDelay(1000);

      mDelay(4000);              // simulate a cycle
      Spindle:= 0;

      if MillCtrlTx(mdStart, '') then
        repeat
          mDelay(10);
        until MillCtrlRx(false);
        if rxMsg = mdStop then
          // handle stopp condition
          HandleStopp;
          return;
        endif;
      else
        // error handling
      endif;
      mDelay(1000);
      MillCtrlRx(false);
    until false;
  endloop;
end;

{--------------------------------------------------------------}
{ Main Program }
{$IDATA}

begin
  CtrlTxRec.msg:= mdWork;
  CtrlTxRec.line:= $11;
  CtrlTxRec.column:= $22;
  CtrlTxRec.param:= $AABB;
  CtrlTxRec.st:= 'test';

  InitPorts;
  TWIrxClear;
  TWItxClear;
  TickTimerTime(1000);
  EnableInts;
  
{$IFDEF LCD}
  LCDsetup_M(LCD_m1);
  LCDcursor_M(LCD_m1, false, false);
  Write(LCDOut_M,  'E-LAB MillBase 3x');
  LCDxy_M(LCD_m1, 0, 1);
  Write(LCDOut_M,  'Test/Demo program');
{$ENDIF}

  mDelay(2000);
  SetMachState([tmInvSetup]);
  SwitchPort1_Clear;

  // test
//  loop
//    TestMillControlDevice;
//  endloop;
  // end test
  
  loop
    SetSwitches;
    if TempTimer > 100 then
      TempTimer:= 0;
      if TWIstat(LM75) then
        TWIout(LM75, 0);                              // prepare read register 0 = temperatur
        TWIinp(LM75, ww);
        MachTemp:= byte(ww);
      else
        MachTemp:= 25;
      endif;
    else
      inc(TempTimer);
    endif;
    
    // Limit Switch active ??
    if [taLimSw] in AxeState then
      TickTimerStop;
      if BuffCount <> 0 then
        FillBlock(@StepBuffer, sizeOf(StepBuffer), 0);
        BuffInpPtr:= 0;
        BuffOutPtr:= 0;
        BuffCount:= 0;
        FeedBusy:= false;
      endif;
      FeedStepsX:= 0;
      FeedStepsY:= 0;
      FeedStepsZ:= 0;
    endif;
    
    if [tmPanicStop] in MachState then
      SwitchPort1_Clear;
    endif;
    
    if not FeedBusy then
      if (BuffCount <> 0) or (FeedStepsX <> 0) or (FeedStepsY <> 0) or (FeedStepsZ <> 0) then
        if FeedStepsZ <> 0 then
          TickTimerTime(FeedSpeedZ);
        else
          TickTimerTime(FeedSpeedX);
        endif;
        FeedBusy:= true;
        TickTimerStart;
      endif;
    elsif BuffXOFF and (BuffCount < XONlim) then
      BuffXOFF:= false;
      SerOut(XON);
    endif;

    if SerStat then
      case SerInp of
        ESC  : // Setup parameters
               case SerInp of
                 'S' : RecvStartSpeed;
                     |
                 'A' : RecvAcceleration;
                     |
                 'M' : RecvMaxSpeed;
                     |
               endcase;
             |
        FS   : RecvFeedSpeed;
             |
        HT   : RequestAxeState;
             |
               // Housekeeping commands
        SYN  : RecvSync;
             |
        ETB  : RequestTemperature;
             |
        ETX  : RequestAbsPosition;
             |
        ENQ  : RequestMachineState;
             |
               // Immediate commands
        BEL  : RecvPanicStop;
             |
        XOFF : RecvStopMove;
             |
        GS   : RecvFlushBuffer;
             |
        STX  : RecvDrivePWR;
             |
        SOH  : RecvSpindleRPM;
             |
        SI   : RecvCoolingOnOff;
             |
               // Move commands
        DC4  : RecvJogSpeed;
             |
        EOT  : RecvFastMoveXYZ;
             |
        FF   : RecvFastMoveXY;
             |
        BS   : RecvFastMoveZ;
             |
               // These Feed cmds respond immediatrely with an <ACK>
               // The HOST must poll the AXES state until all steps are executed
        EM   : RecvFeedStepsX;
             |
        SUB  : RecvFeedStepsY;
             |
        VT   : RecvFeedStepsZ;
             |
        RS   : RecvClearPosition;
             |
               // This Feed cmd doesn't respond at all
               // After Receiving all consecute cmds the EndOfFeed must be received
        DC2  : RecvFeedStepsXYZ;
             |
               // the PC-master sends this request after a block of feed steps XYZ
               // it notices the controller that the block transfer is finished now
        CR   : RecvEndOfFeed;
             |
        LF   : RequestStart;
             |
        SO   : RecvLimSwitchPol;
             |
        DLE  : RecvRefSwitchPol;
             |
        CAN  : RecvVelocityXYZ;
             |
        US   : RecvRefMoveXYZ;
             |
        NAK  :
             |
        XON  :
             |
      endcase;
    endif;
  endloop;
(*
    // function TWItxBroadcast(cmd : byte; subnode : byte [; count : byte]) : boolean;
    TWIBROADCASTBUFFER[0]:= bb+1;
    TWIBROADCASTBUFFER[1]:= bb+2;
    TWIBROADCASTBUFFER[2]:= bb+3;
    TWIBROADCASTBUFFER[3]:= bb+4;
    bool:= TWItxBroadcast($AA, 0, 4);

    TWIrxClear;
    bool:= TWIrxFrame(TWIslvX);

    TWItxClear;
    TWItxBuff[0]:= $AA;
    TWItxBuff[1]:= $BB;
    TWItxBuff[2]:= $CC;
    TWItxBuff[3]:= $DD;
    bool:= TWItxFrame(TWIslvX, 4);
*)
end MillBase3x.

